#!/usr/bin/env python3

import os, sys, shutil
import numpy as np
from scipy.interpolate import CubicSpline

class TabInter (object):
    def __init__(self,
                 filename):
        self.reinit(filename)
        
    def reinit(self,
               filename):
        self.vdata = np.loadtxt(filename)
        self.rmin = self.vdata[0][0]
        self.hh = self.vdata[1][0] - self.vdata[0][0]
        self.nspline = self.vdata.shape[0] - 1
        ncol = self.vdata.shape[1] - 1
        n0 = (-1 + np.sqrt(1 + 8 * ncol)) * 0.5
        self.ntypes = int(n0 + 0.1)
        assert(self.ntypes * (self.ntypes+1) // 2 == ncol),\
            "number of volumes provided in %s does not match guessed number of types %d" % (filename, self.ntypes)
        self.tab_info = np.array([self.rmin, self.hh, self.nspline, self.ntypes])
        self.tab_data = self._make_data()

    def get(self) :
        return self.tab_info, self.tab_data

    def _make_data(self) :
        data = np.zeros([self.ntypes * self.ntypes * 4 * self.nspline])
        stride = 4 * self.nspline
        idx_iter = 0
        xx = self.vdata[:,0]
        for t0 in range(self.ntypes) :
            for t1 in range(t0, self.ntypes) :
                vv = self.vdata[:,1+idx_iter]
                cs = CubicSpline(xx, vv)
                dd = cs(xx, 1)
                dd *= self.hh
                dtmp = np.zeros(stride)
                for ii in range(self.nspline) :
                    dtmp[ii*4+0] = 2 * vv[ii] - 2 * vv[ii+1] +     dd[ii] + dd[ii+1]
                    dtmp[ii*4+1] =-3 * vv[ii] + 3 * vv[ii+1] - 2 * dd[ii] - dd[ii+1]
                    dtmp[ii*4+2] = dd[ii]
                    dtmp[ii*4+3] = vv[ii]
                data[(t0 * self.ntypes + t1) * stride : (t0 * self.ntypes + t1) * stride + stride] \
                    = dtmp
                data[(t1 * self.ntypes + t0) * stride : (t1 * self.ntypes + t0) * stride + stride] \
                    = dtmp
                idx_iter += 1
        return data



import os,sys
import numpy as np
import unittest

from deepmd.env import tf
from tensorflow.python.framework import ops

# load grad of force module
import deepmd._prod_force_grad
import deepmd._prod_virial_grad
import deepmd._prod_force_se_r_grad
import deepmd._prod_virial_se_r_grad
import deepmd._soft_min_force_grad
import deepmd._soft_min_virial_grad

from common import force_test
from common import virial_test
from common import force_dw_test
from common import virial_dw_test
from common import Data

from deepmd.DescrptSeAR import DescrptSeAR

from deepmd.RunOptions import global_tf_float_precision
from deepmd.RunOptions import global_np_float_precision
from deepmd.RunOptions import global_ener_float_precision

class Inter():
    def __init__ (self, 
                  data) :
        self.sess = tf.Session()
        self.data = data
        self.natoms = self.data.get_natoms()
        self.ntypes = self.data.get_ntypes()
        param_a = {
            'sel' : [12,24],
            'rcut': 4,
            'rcut_smth' : 3.5,
            'neuron': [5, 10, 20],
            'seed': 1,
        }
        param_r = {
            'sel' : [20,40],
            'rcut': 6,
            'rcut_smth' : 6.5,
            'neuron': [10, 20, 40],
            'seed': 1,
        }
        param = {'a': param_a, 'r': param_r}
        self.descrpt = DescrptSeAR(param)
        self.ndescrpt = self.descrpt.get_dim_out()
        # davg = np.zeros ([self.ntypes, self.ndescrpt])
        # dstd = np.ones  ([self.ntypes, self.ndescrpt])
        # self.t_avg = tf.constant(davg.astype(np.float64))
        # self.t_std = tf.constant(dstd.astype(np.float64))
        avg_a = np.zeros([self.ntypes, self.descrpt.descrpt_a.ndescrpt])
        std_a = np.ones ([self.ntypes, self.descrpt.descrpt_a.ndescrpt])
        avg_r = np.zeros([self.ntypes, self.descrpt.descrpt_r.ndescrpt])
        std_r = np.ones ([self.ntypes, self.descrpt.descrpt_r.ndescrpt])
        self.avg = [avg_a, avg_r]
        self.std = [std_a, std_r]
        self.default_mesh = np.zeros (6, dtype = np.int32)
        self.default_mesh[3] = 2
        self.default_mesh[4] = 2
        self.default_mesh[5] = 2
        # make place holder
        self.coord      = tf.placeholder(global_tf_float_precision, [None, self.natoms[0] * 3], name='t_coord')
        self.box        = tf.placeholder(global_tf_float_precision, [None, 9], name='t_box')
        self.type       = tf.placeholder(tf.int32,   [None, self.natoms[0]], name = "t_type")
        self.tnatoms    = tf.placeholder(tf.int32,   [None], name = "t_natoms")
        
    def _net (self,
              inputs, 
              name,
              reuse = False) :
        with tf.variable_scope(name, reuse=reuse):
            net_w = tf.get_variable ('net_w', 
                                     [self.descrpt.get_dim_out()], 
                                     global_tf_float_precision,
                                     tf.constant_initializer (self.net_w_i))
        dot_v = tf.matmul (tf.reshape (inputs, [-1, self.descrpt.get_dim_out()]),
                           tf.reshape (net_w, [self.descrpt.get_dim_out(), 1]))
        return tf.reshape (dot_v, [-1])
        
    def comp_ef (self, 
                 dcoord, 
                 dbox, 
                 dtype,
                 tnatoms,
                 name,
                 reuse = None) :
        dout = self.descrpt.build(dcoord, dtype, tnatoms, dbox, self.default_mesh, self.avg, self.std, suffix=name, reuse=reuse)
        inputs_reshape = tf.reshape (dout, [-1, self.descrpt.get_dim_out()])
        atom_ener = self._net (inputs_reshape, name, reuse = reuse)
        atom_ener_reshape = tf.reshape(atom_ener, [-1, self.natoms[0]])       
        energy = tf.reduce_sum (atom_ener_reshape, axis = 1)        
        force, virial, av = self.descrpt.prod_force_virial(atom_ener_reshape, tnatoms)
        return energy, force, virial


class TestDescrptAR(Inter, unittest.TestCase):
    def __init__ (self, *args, **kwargs):
        data = Data()
        Inter.__init__(self, data)
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.controller = object()

    def test_force (self) :
        force_test(self, self, suffix = '_se_ar')

    def test_virial (self) :
        virial_test(self, self, suffix = '_se_ar')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
import platform
import os,sys
import numpy as np

from deepmd.env import tf

from tensorflow.python.framework import ops

if platform.system() == "Windows":
    ext = "dll"
elif platform.system() == "Darwin":
    ext = "dylib"
else:
    ext = "so"
module_path = os.path.dirname(os.path.realpath(__file__))
assert (os.path.isfile (os.path.join(module_path, "libop_abi.{}".format(ext)))), "op module does not exist"
op_module = tf.load_op_library(os.path.join(module_path, "libop_abi.{}".format(ext)))

class DeepEval():
    """
    common methods for DeepPot, DeepWFC, DeepPolar, ...
    """
    def __init__(self, 
                 model_file) :
        model_file = model_file
        graph = self.load_graph (model_file)
        t_mt = graph.get_tensor_by_name('load/model_attr/model_type:0')
        sess = tf.Session (graph = graph)        
        [mt] = sess.run([t_mt], feed_dict = {})
        self.model_type = mt.decode('utf-8')

    def load_graph(self, 
                   frozen_graph_filename, 
                   prefix = 'load'):
        # We load the protobuf file from the disk and parse it to retrieve the 
        # unserialized graph_def
        with tf.gfile.GFile(frozen_graph_filename, "rb") as f:
            graph_def = tf.GraphDef()
            graph_def.ParseFromString(f.read())

        # Then, we can use again a convenient built-in function to import a graph_def into the 
        # current default Graph
        with tf.Graph().as_default() as graph:
            tf.import_graph_def(
                graph_def, 
                input_map=None, 
                return_elements=None, 
                name=prefix, 
                producer_op_list=None
            )
        return graph

    def make_default_mesh(self, test_box) :
        ncell = np.ones (3, dtype=np.int32)
        avg_box = np.average (test_box, axis = 0)
        cell_size = 3
        avg_box = np.reshape (avg_box, [3,3])
        for ii in range (3) :
            ncell[ii] = int ( np.linalg.norm(avg_box[ii]) / cell_size )
            if (ncell[ii] < 2) : ncell[ii] = 2
        default_mesh = np.zeros (6, dtype = np.int32)
        default_mesh[3] = ncell[0]
        default_mesh[4] = ncell[1]
        default_mesh[5] = ncell[2]
        return default_mesh

    def sort_input(self, coord, atom_type, sel_atoms = None) :
        if sel_atoms is not None:
            selection = [False] * np.size(atom_type)
            for ii in sel_atoms:
                selection += (atom_type == ii)
            sel_atom_type = atom_type[selection]
        natoms = atom_type.size
        idx = np.arange (natoms)
        idx_map = np.lexsort ((idx, atom_type))
        nframes = coord.shape[0]
        coord = coord.reshape([nframes, -1, 3])
        coord = np.reshape(coord[:,idx_map,:], [nframes, -1])
        atom_type = atom_type[idx_map]
        if sel_atoms is not None:
            sel_natoms = np.size(sel_atom_type)
            sel_idx = np.arange(sel_natoms)
            sel_idx_map = np.lexsort((sel_idx, sel_atom_type))
            sel_atom_type = sel_atom_type[sel_idx_map]
            return coord, atom_type, idx_map, sel_atom_type, sel_idx_map
        else:
            return coord, atom_type, idx_map

    def reverse_map(self, vec, imap):
        ret = np.zeros(vec.shape)        
        for idx,ii in enumerate(imap) :
            ret[:,ii,:] = vec[:,idx,:]
        return ret
        
    def make_natoms_vec(self, atom_types) :
        natoms_vec = np.zeros (self.ntypes+2).astype(int)
        natoms = atom_types.size
        natoms_vec[0] = natoms
        natoms_vec[1] = natoms
        for ii in range (self.ntypes) :
            natoms_vec[ii+2] = np.count_nonzero(atom_types == ii)
        return natoms_vec


class DeepTensor(DeepEval) :
    """
    Evaluates a tensor model
    """
    def __init__(self, 
                 model_file, 
                 variable_name,                  
                 variable_dof) :
        DeepEval.__init__(self, model_file)
        self.model_file = model_file
        self.graph = self.load_graph (self.model_file)
        self.variable_name = variable_name
        self.variable_dof = variable_dof
        # checkout input/output tensors from graph
        self.t_ntypes = self.graph.get_tensor_by_name ('load/descrpt_attr/ntypes:0')
        self.t_rcut   = self.graph.get_tensor_by_name ('load/descrpt_attr/rcut:0')
        self.t_tmap   = self.graph.get_tensor_by_name ('load/model_attr/tmap:0')
        self.t_sel_type= self.graph.get_tensor_by_name ('load/model_attr/sel_type:0')
        # inputs
        self.t_coord  = self.graph.get_tensor_by_name ('load/t_coord:0')
        self.t_type   = self.graph.get_tensor_by_name ('load/t_type:0')
        self.t_natoms = self.graph.get_tensor_by_name ('load/t_natoms:0')
        self.t_box    = self.graph.get_tensor_by_name ('load/t_box:0')
        self.t_mesh   = self.graph.get_tensor_by_name ('load/t_mesh:0')
        # outputs
        self.t_tensor = self.graph.get_tensor_by_name ('load/o_%s:0' % self.variable_name)
        # start a tf session associated to the graph
        self.sess = tf.Session (graph = self.graph)        
        [self.ntypes, self.rcut, self.tmap, self.tselt] = self.sess.run([self.t_ntypes, self.t_rcut, self.t_tmap, self.t_sel_type])
        self.tmap = self.tmap.decode('UTF-8').split()

    def get_ntypes(self) :
        return self.ntypes

    def get_rcut(self) :
        return self.rcut

    def get_type_map(self):
        return self.tmap

    def get_sel_type(self):
        return self.tselt

    def eval(self,
             coords, 
             cells, 
             atom_types) :
        # standarize the shape of inputs
        coords = np.array(coords)
        cells = np.array(cells)
        atom_types = np.array(atom_types, dtype = int)

        # reshape the inputs 
        cells = np.reshape(cells, [-1, 9])
        nframes = cells.shape[0]
        coords = np.reshape(coords, [nframes, -1])
        natoms = coords.shape[1] // 3

        # sort inputs
        coords, atom_types, imap, sel_at, sel_imap = self.sort_input(coords, atom_types, sel_atoms = self.get_sel_type())

        # make natoms_vec and default_mesh
        natoms_vec = self.make_natoms_vec(atom_types)
        assert(natoms_vec[0] == natoms)
        default_mesh = self.make_default_mesh(cells)

        # evaluate
        tensor = []
        feed_dict_test = {}
        feed_dict_test[self.t_natoms] = natoms_vec
        feed_dict_test[self.t_mesh  ] = default_mesh
        feed_dict_test[self.t_type  ] = atom_types
        t_out = [self.t_tensor]
        for ii in range(nframes) :
            feed_dict_test[self.t_coord] = np.reshape(coords[ii:ii+1, :], [-1])
            feed_dict_test[self.t_box  ] = np.reshape(cells [ii:ii+1, :], [-1])
            v_out = self.sess.run (t_out, feed_dict = feed_dict_test)
            tensor.append(v_out[0])

        # reverse map of the outputs
        tensor = np.array(tensor)
        tensor = self.reverse_map(np.reshape(tensor, [nframes,-1,self.variable_dof]), sel_imap)

        tensor = np.reshape(tensor, [nframes, len(sel_at), self.variable_dof])
        return tensor

    



import os,sys
import numpy as np
import unittest

from deepmd.env import tf
from tensorflow.python.framework import ops

# load grad of force module
import deepmd._prod_force_grad
import deepmd._prod_virial_grad
import deepmd._soft_min_force_grad
import deepmd._soft_min_virial_grad

from common import force_test
from common import virial_test
from common import force_dw_test
from common import virial_dw_test
from common import Data

from deepmd.DescrptLocFrame import op_module

from deepmd.RunOptions import global_tf_float_precision
from deepmd.RunOptions import global_np_float_precision
from deepmd.RunOptions import global_ener_float_precision

class Inter():
    def __init__ (self,
                  data,
                  comp = 0) :
        self.sess = tf.Session()
        self.data = data
        self.natoms = self.data.get_natoms()
        self.ntypes = self.data.get_ntypes()
        self.sel_a = [12,24]
        self.sel_r = [12,24]
        self.rcut_a = -1
        self.rcut_r = 10.0
        self.axis_rule = [0, 1, 0, 0, 1, 1, 0, 0, 0, 0, 1, 0]
        self.nnei_a = np.cumsum(self.sel_a)[-1]
        self.nnei_r = np.cumsum(self.sel_r)[-1]
        self.nnei = self.nnei_a + self.nnei_r
        self.ndescrpt_a = self.nnei_a * 4
        self.ndescrpt_r = self.nnei_r * 1
        self.ndescrpt = self.ndescrpt_a + self.ndescrpt_r
        davg = np.zeros ([self.ntypes, self.ndescrpt])
        dstd = np.ones  ([self.ntypes, self.ndescrpt])
        self.t_avg = tf.constant(davg.astype(global_np_float_precision))
        self.t_std = tf.constant(dstd.astype(global_np_float_precision))
        self.default_mesh = np.zeros (6, dtype = np.int32)
        self.default_mesh[3] = 2
        self.default_mesh[4] = 2
        self.default_mesh[5] = 2
        # make place holder
        self.coord      = tf.placeholder(global_tf_float_precision, [None, self.natoms[0] * 3], name='t_coord')
        self.box        = tf.placeholder(global_tf_float_precision, [None, 9], name='t_box')
        self.type       = tf.placeholder(tf.int32,   [None, self.natoms[0]], name = "t_type")
        self.tnatoms    = tf.placeholder(tf.int32,   [None], name = "t_natoms")

        
    def _net (self,
             inputs, 
             name,
             reuse = False) :
        with tf.variable_scope(name, reuse=reuse):
            net_w = tf.get_variable ('net_w', 
                                     [self.ndescrpt], 
                                     global_tf_float_precision,
                                     tf.constant_initializer (self.net_w_i))
        dot_v = tf.matmul (tf.reshape (inputs, [-1, self.ndescrpt]),
                           tf.reshape (net_w, [self.ndescrpt, 1]))
        return tf.reshape (dot_v, [-1])
        
    def comp_ef (self, 
                 dcoord, 
                 dbox, 
                 dtype,
                 tnatoms,
                 name,
                 reuse = None) :
        descrpt, descrpt_deriv, rij, nlist, axis, rot_mat \
            = op_module.descrpt (dcoord, 
                                 dtype,
                                 tnatoms,
                                 dbox, 
                                 tf.constant(self.default_mesh),
                                 self.t_avg,
                                 self.t_std,
                                 rcut_a = self.rcut_a, 
                                 rcut_r = self.rcut_r, 
                                 sel_a = self.sel_a, 
                                 sel_r = self.sel_r, 
                                 axis_rule = self.axis_rule)
        self.axis = axis
        self.nlist = nlist
        self.descrpt = descrpt
        inputs_reshape = tf.reshape (descrpt, [-1, self.ndescrpt])
        atom_ener = self._net (inputs_reshape, name, reuse = reuse)
        atom_ener_reshape = tf.reshape(atom_ener, [-1, self.natoms[0]])        
        energy = tf.reduce_sum (atom_ener_reshape, axis = 1)        
        net_deriv_ = tf.gradients (atom_ener, inputs_reshape)
        net_deriv = net_deriv_[0]
        net_deriv_reshape = tf.reshape (net_deriv, [-1, self.natoms[0] * self.ndescrpt]) 

        force = op_module.prod_force (net_deriv_reshape, 
                                      descrpt_deriv, 
                                      nlist, 
                                      axis, 
                                      tnatoms,
                                      n_a_sel = self.nnei_a, 
                                      n_r_sel = self.nnei_r)
        virial, atom_vir = op_module.prod_virial (net_deriv_reshape, 
                                                  descrpt_deriv, 
                                                  rij,
                                                  nlist, 
                                                  axis, 
                                                  tnatoms,
                                                  n_a_sel = self.nnei_a, 
                                                  n_r_sel = self.nnei_r)
        return energy, force, virial


    def comp_f_dw (self, 
                   dcoord, 
                   dbox, 
                   dtype,                 
                   tnatoms,
                   name,
                   reuse = None) :
        energy, force, virial = self.comp_ef (dcoord, dbox, dtype, tnatoms, name, reuse)
        with tf.variable_scope(name, reuse=True):
            net_w = tf.get_variable ('net_w', [self.ndescrpt], global_tf_float_precision, tf.constant_initializer (self.net_w_i))
        f_mag = tf.reduce_sum (tf.nn.tanh(force))
        f_mag_dw = tf.gradients (f_mag, net_w)
        assert (len(f_mag_dw) == 1), "length of dw is wrong"        
        return f_mag, f_mag_dw[0]


    def comp_v_dw (self, 
                   dcoord, 
                   dbox, 
                   dtype,                 
                   tnatoms,
                   name,
                   reuse = None) :
        energy, force, virial = self.comp_ef (dcoord, dbox, dtype, tnatoms, name, reuse)
        with tf.variable_scope(name, reuse=True):
            net_w = tf.get_variable ('net_w', [self.ndescrpt], global_tf_float_precision, tf.constant_initializer (self.net_w_i))
        v_mag = tf.reduce_sum (virial)
        v_mag_dw = tf.gradients (v_mag, net_w)
        assert (len(v_mag_dw) == 1), "length of dw is wrong"        
        return v_mag, v_mag_dw[0]



class TestNonSmooth(Inter, unittest.TestCase):
    def __init__ (self, *args, **kwargs):
        self.places = 5
        data = Data()
        Inter.__init__(self, data)
        unittest.TestCase.__init__(self, *args, **kwargs)
        self.controller = object()

    def test_force (self) :
        force_test(self, self)
        # t_energy, t_force, t_virial \
        #     = self.comp_ef (self.coord, self.box, self.type, self.tnatoms, name = "test")
        # self.sess.run (tf.global_variables_initializer())
        # dcoord, dbox, dtype = self.data.get_data ()
        # hh = 1e-6
        # dcoordp = np.copy(dcoord)
        # dcoordm = np.copy(dcoord)
        # dcoordp[0,0] = dcoord[0,0] + hh
        # dcoordm[0,0] = dcoord[0,0] - hh
        # [axis0, nlist0, d0] = self.sess.run ([self.axis, self.nlist, self.descrpt], 
        #                                  feed_dict = {
        #                                      self.coord:     dcoordp,
        #                                      self.box:       dbox,
        #                                      self.type:      dtype,
        #                                      self.tnatoms:   self.natoms}
        # )
        # [axis1, nlist1, d1] = self.sess.run ([self.axis, self.nlist, self.descrpt], 
        #                                  feed_dict = {
        #                                      self.coord:     dcoordm,
        #                                      self.box:       dbox,
        #                                      self.type:      dtype,
        #                                      self.tnatoms:   self.natoms}
        # )
        # print((nlist0 - nlist1))
        # print((axis0 - axis1))

    def test_virial (self) :
        virial_test(self, self)

    def test_force_dw (self) :
        force_dw_test(self, self)

    def test_virial_dw (self) :
        virial_dw_test(self, self)


if __name__ == '__main__':
    unittest.main()
